@@ -16,6 +16,7 @@ use Debconf::Question;
 use Debconf::Priority qw(priority_valid high_enough);
 use Debconf::FrontEnd::Noninteractive;
 use Debconf::Log ':all';
+use Debconf::Encoding;
 use base qw(Debconf::Base);
 
 =head1 DESCRIPTION
@@ -87,11 +88,13 @@ An array reference. If set, it will hold a list of named of question that are
 # Here I define all the numeric result codes that are used.
 my %codes = (
 	success => 0,
+	escaped_data => 1,
 	badparams => 10,
 	syntaxerror => 20,
 	input_invisible => 30,
 	version_bad => 30,
 	go_back => 30,
+	progresscancel => 30,
 	internalerror => 100,
 );
 
@@ -136,7 +139,7 @@ sub startup {
 	my $confmodule=shift;
 
 	# There is an implicit clearing of any previously pending questions
-	# when a new confmodule is ran.
+	# when a new confmodule is run.
 	$this->frontend->clear;
 	$this->busy([]);
 	
@@ -175,6 +178,46 @@ sub communicate {
 	return 1;
 }
 
+=item escape
+
+Escape backslashes and newlines for output via the debconf protocol.
+
+=cut
+
+sub escape {
+	my $text=shift;
+	$text=~s/\\/\\\\/g;
+	$text=~s/\n/\\n/g;
+	return $text;
+}
+
+=item unescape_split
+
+Unescape text received via the debconf protocol, and split by unescaped
+whitespace.
+
+=cut
+
+sub unescape_split {
+	my $text=shift;
+	my @words;
+	my $word='';
+	for my $chunk (split /(\\.|\s+)/, $text) {
+		if ($chunk eq '\n') {
+			$word.="\n";
+		} elsif ($chunk=~/^\\(.)$/) {
+			$word.=$1;
+		} elsif ($chunk=~/^\s+$/) {
+			push @words, $word;
+			$word='';
+		} else {
+			$word.=$chunk;
+		}
+	}
+	push @words, $word if $word ne '';
+	return @words;
+}
+
 =item process_command
 
 Pass in a raw command, and it will be processed and handled.
@@ -187,7 +230,12 @@ sub process_command {
 	debug developer => "<-- $_";
 	return 1 unless defined && ! /^\s*#/; # Skip blank lines, comments.
 	chomp;
-	my ($command, @params)=split(' ', $_);
+	my ($command, @params);
+	if (defined $this->client_capb and grep { $_ eq 'escape' } @{$this->client_capb}) {
+		($command, @params)=unescape_split($_);
+	} else {
+		($command, @params)=split(' ', $_);
+	}
 	$command=lc($command);
 	# This command could not be handled by a sub.
 	if (lc($command) eq "stop") {
@@ -204,7 +252,7 @@ sub process_command {
 	debug developer => "--> $ret";
 	if ($ret=~/\n/) {
 		debug developer => 'Warning: return value is multiline, and would break the debconf protocol. Truncating to first line.';
-		$ret=~s/\n.*//;
+		$ret=~s/\n.*//s;
 		debug developer => "--> $ret";
 	}
 	return $ret;
@@ -265,15 +313,26 @@ sub command_input {
 	# not.
 	my $visible=1;
 
+	# Error questions are always shown even if they're asked at a low
+	# priority or have already been seen.
+	if ($question->type ne 'error') {
+		# Don't show items that are unimportant.
+		$visible='' unless high_enough($priority);
+
+		# Don't re-show already seen questions, unless reconfiguring.
+		$visible='' if ! Debconf::Config->reshow &&
+			       $question->flag('seen') eq 'true';
+	}
+
+	# We may want to set the seen flag on noninteractive questions
+	# even though they aren't shown.
+	my $markseen=$visible;
+
 	# Noninteractive frontends never show anything.
-	$visible='' if ! $this->frontend->interactive;
-
-	# Don't show items that are unimportant.
-	$visible='' unless high_enough($priority);
-
-	# Don't re-show already seen questions, unless reconfiguring.
-	$visible='' if ! Debconf::Config->reshow &&
-	               $question->flag('seen') eq 'true';
+	if ($visible && ! $this->frontend->interactive) {
+		$visible='';
+		$markseen='' unless Debconf::Config->noninteractive_seen eq 'true';
+	}
 
 	my $element;
 	if ($visible) {
@@ -304,6 +363,8 @@ sub command_input {
 		return $codes{input_invisible}, "question skipped" unless $element;
 	}
 
+	$element->markseen($markseen);
+
 	push @{$this->busy}, $question_name;
 	
 	$this->frontend->add($element);
@@ -363,10 +424,14 @@ sub command_capb {
 	my $this=shift;
 	$this->client_capb([@_]);
 	# Set capb_backup on the frontend if the client can backup.
-	$this->frontend->capb_backup(1) if grep { $_ eq 'backup' } @_;
+	if (grep { $_ eq 'backup' } @_) {
+		$this->frontend->capb_backup(1);
+	} else {
+		$this->frontend->capb_backup('');
+	}
 	# Multiselect is added as a capability to fix a backwards
 	# compatability problem.
-	my @capb=('multiselect');
+	my @capb=('multiselect', 'escape');
 	push @capb, $this->frontend->capb;
 	return $codes{success}, @capb;
 }
@@ -380,6 +445,7 @@ Stores the specified title in the associated FrontEnds title field.
 sub command_title {
 	my $this=shift;
 	$this->frontend->title(join ' ', @_);
+	$this->frontend->requested_title($this->frontend->title);
 
 	return $codes{success};
 }
@@ -399,7 +465,12 @@ sub command_settitle {
 	my $question=Debconf::Question->get($question_name) ||
 		return $codes{badparams}, "\"$question_name\" doesn't exist";
 
-	$this->frontend->title($question->description);
+	if ($this->frontend->can('settitle')) {
+		$this->frontend->settitle($question);
+	} else {
+		$this->frontend->title($question->description);
+	}
+	$this->frontend->requested_title($this->frontend->title);
 	
 	return $codes{success};
 }
@@ -438,7 +509,7 @@ sub command_go {
 	             grep { $_->visible } @{$this->frontend->elements})) {
 		foreach (@{$this->frontend->elements}) {
 			$_->question->value($_->value);
-			push @{$this->seen}, $_->question if $_->visible && $_->question;
+			push @{$this->seen}, $_->question if $_->markseen && $_->question;
 		}
 		$this->frontend->clear;
 		$this->busy([]);
@@ -467,8 +538,13 @@ sub command_get {
 	my $question=Debconf::Question->get($question_name) ||
 		return $codes{badparams}, "$question_name doesn't exist";
 
-	if (defined $question->value) {
-		return $codes{success}, $question->value;
+	my $value=$question->value;
+	if (defined $value) {
+		if (defined $this->client_capb and grep { $_ eq 'escape' } @{$this->client_capb}) {
+			return $codes{escaped_data}, escape($value);
+		} else {
+			return $codes{success}, $value;
+		}
 	}
 	else {
 		return $codes{success}, '';
@@ -619,11 +695,16 @@ sub command_metaget {
 	
 	my $question=Debconf::Question->get($question_name) ||
 		return $codes{badparams}, "$question_name doesn't exist";
-	my $fieldval=$question->$field();
+	my $lcfield=lc $field;
+	my $fieldval=$question->$lcfield();
 	unless (defined $fieldval) {
 		return $codes{badparams}, "$field does not exist";
 	}
-	return $codes{success}, $fieldval;
+	if (defined $this->client_capb and grep { $_ eq 'escape' } @{$this->client_capb}) {
+		return $codes{escaped_data}, escape($fieldval);
+	} else {
+		return $codes{success}, $fieldval;
+	}
 }
 
 =item command_fget
@@ -675,6 +756,171 @@ sub command_fset {
 	return $codes{success}, $question->flag($flag, $value);
 }
 
+=item command_info
+
+Pass this a question name. It displays the given template as an out-of-band
+informative message. Unlike inputting a note, this doesn't require an
+acknowledgement from the user, and depending on the frontend it may not even
+be displayed at all. Frontends should display the info persistently until
+some other info comes along.
+
+With no arguments, this resets the info message to a default value.
+
+=cut
+
+sub command_info {
+	my $this=shift;
+
+	if (@_ == 0) {
+		$this->frontend->info(undef);
+	} elsif (@_ == 1) {
+		my $question_name=shift;
+
+		my $question=Debconf::Question->get($question_name) ||
+			return $codes{badparams}, "\"$question_name\" doesn't exist";
+
+		$this->frontend->info($question);
+	} else {
+		return $codes{syntaxerror}, "Incorrect number of arguments";
+	}
+
+	return $codes{success};
+}
+
+=item command_progress
+
+Progress bar handling. Pass this a subcommand name followed by any arguments
+required by the subcommand, as follows:
+
+=over 4
+
+=item START
+
+Pass this a minimum value, a maximum value, and a question name. It creates
+a progress bar with the specified range and the description of the specified
+question as the title.
+
+=item SET
+
+Pass this a value. It sets the current position of the progress bar to the
+specified value.
+
+=item STEP
+
+Pass this an increment. It increments the current position of the progress
+bar by the specified amount.
+
+=item INFO
+
+Pass this a template name. It displays the specified template as an
+informational message in the progress bar.
+
+=item STOP
+
+This subcommand takes no arguments. It destroys the progress bar.
+
+=back
+
+Note that the frontend's progress_set, progress_step, and progress_info 
+functions should return true, unless the progress bar was canceled.
+
+=cut
+
+sub command_progress {
+	my $this=shift;
+	return $codes{syntaxerror}, "Incorrect number of arguments" if @_ < 1;
+	my $subcommand=shift;
+	$subcommand=lc($subcommand);
+	
+	my $ret;
+
+	if ($subcommand eq 'start') {
+		return $codes{syntaxerror}, "Incorrect number of arguments" if @_ != 3;
+		my $min=shift;
+		my $max=shift;
+		my $question_name=shift;
+
+		return $codes{syntaxerror}, "min ($min) > max ($max)" if $min > $max;
+
+		my $question=Debconf::Question->get($question_name) ||
+			return $codes{badparams}, "$question_name doesn't exist";
+
+		$this->frontend->progress_start($min, $max, $question);
+		$ret=1;
+	}
+	elsif ($subcommand eq 'set') {
+		return $codes{syntaxerror}, "Incorrect number of arguments" if @_ != 1;
+		my $value=shift;
+		$ret = $this->frontend->progress_set($value);
+	}
+	elsif ($subcommand eq 'step') {
+		return $codes{syntaxerror}, "Incorrect number of arguments" if @_ != 1;
+		my $inc=shift;
+		$ret = $this->frontend->progress_step($inc);
+	}
+	elsif ($subcommand eq 'info') {
+		return $codes{syntaxerror}, "Incorrect number of arguments" if @_ != 1;
+		my $question_name=shift;
+
+		my $question=Debconf::Question->get($question_name) ||
+			return $codes{badparams}, "$question_name doesn't exist";
+
+		$ret = $this->frontend->progress_info($question);
+	}
+	elsif ($subcommand eq 'stop') {
+		return $codes{syntaxerror}, "Incorrect number of arguments" if @_ != 0;
+		$this->frontend->progress_stop();
+		$ret=1;
+	}
+	else {
+		return $codes{syntaxerror}, "Unknown subcommand";
+	}
+
+	if ($ret) {
+		return $codes{success}, "OK";
+	}
+	else {
+		return $codes{progresscancel}, "CANCELED";
+	}
+}
+
+=item command_data
+
+Accept template data from the client, for use on the UI agent side of the
+passthrough frontend.
+
+TODO: Since process_command() collapses multiple spaces in commands into
+single spaces, this doesn't quite handle bulleted lists correctly.
+
+=cut
+
+sub command_data {
+	my $this=shift;
+	return $codes{syntaxerror}, "Incorrect number of arguments" if @_ < 3;
+	my $template=shift;
+	my $item=shift;
+	my $value=join(' ', @_);
+	$value=~s/\\([n"\\])/($1 eq 'n') ? "\n" : $1/eg;
+
+	my $tempobj=Debconf::Template->get($template);
+	if (! $tempobj) {
+		if ($item ne 'type') {
+			return $codes{badparams}, "Template data field '$item' received before type field";
+		}
+		$tempobj=Debconf::Template->new($template, $this->owner, $value);
+		if (! $tempobj) {
+			return $codes{internalerror}, "Internal error making template";
+		}
+	} else {
+		if ($item eq 'type') {
+			return $codes{badparams}, "Template type already set";
+		}
+		$tempobj->$item(Debconf::Encoding::convert("UTF-8", $value));
+	}
+
+	return $codes{success};
+}
+
 =item command_visible
 
 Deprecated.
@@ -707,6 +953,38 @@ sub command_exist {
 		Debconf::Question->get($question_name) ? "true" : "false";
 }
 
+=item command_x_loadtemplatefile
+
+Extension to load a specified template file.
+
+=cut
+
+sub command_x_loadtemplatefile {
+	my $this=shift;
+
+	return $codes{syntaxerror}, "Incorrect number of arguments" if @_ < 1 || @_ > 2;
+
+	my $file=shift;
+	my $fh=FileHandle->new($file);
+	if (! $fh) {
+		return $codes{badparams}, "failed to open $file: $!";
+	}
+
+	my $owner=$this->owner;
+	if (@_) {
+		$owner=shift;
+	}
+
+	eval {
+		Debconf::Template->load($fh, $owner);
+	};
+	if ($@) {
+		$@=~s/\n/\\n/g;
+		return $codes{internalerror}, $@;
+	}
+	return $codes{success};
+}
+
 =item AUTOLOAD
 
 Handles storing and loading fields.
